"""OpenAI embedding"""

import logging
import openai
from openai.embeddings_utils import get_embedding
from typing import Dict, List
from chatdbt.model import EmbeddingProvider
from tenacity import retry, stop_after_attempt, wait_random_exponential


COMPLETION_MODEL = "gpt-3.5-turbo"
EMBEDDING_MODEL = "text-embedding-ada-002"


@retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(6))
def chat_completion(messages: List[Dict[str, str]], temperature=0.2):
    return openai.ChatCompletion.create(
        messages=messages, model=COMPLETION_MODEL, temperature=temperature
    )


def price_for_completion(n_tokens: int) -> float:
    return float(n_tokens) / 1000.0 * 0.002


def price_for_embedding(n_tokens: int) -> float:
    return float(n_tokens) / 1000.0 * 0.0004


class Openai(EmbeddingProvider):
    """OpenAI embedding"""

    def __init__(self, temperature: float = 0.2):
        self.embedding_model = EMBEDDING_MODEL
        self.temperature = float(temperature)

    def embed(self, content: str) -> List[float]:
        """Embed a piece of text into a vector"""
        return get_embedding(content, engine=self.embedding_model)

    def completion(self):
        openai.ChatCompletion.create()

    def chat_completion(self, messages: List[Dict[str, str]]) -> str:
        res = chat_completion(messages, temperature=self.temperature)
        usage = res["usage"]["total_tokens"]
        logging.info(
            "chat-completion total tokens: %s, cost %s$",
            usage,
            price_for_completion(usage),
        )

        return res["choices"][0]["message"]["content"]



import logging
from typing import Optional, Set, List, Dict, cast, Any
import uuid
import datetime
from chatdbt.model import (
    ChatConversationDocument,
    ChatMessageMeta,
    Doc,
    DocMetaContainer,
    DocType,
    DBTDocResolver,
    VectorStorage,
    ChatMessage,
    TikTokenProvider,
    DBTModelDocument,
    DBTModelSqlDocument,
    DBTDocMeta,
    CatalogColumn,
)
from collections import defaultdict
from chatdbt.openai import (
    Openai,
    price_for_embedding,
    EMBEDDING_MODEL,
)
from chatdbt.i18n import get_i18n_text, I18nKey


def _truncate_schema_name_for_model(name: str) -> str:
    """Truncate the schema name from a model name"""
    return ".".join(name.split(".")[1:])


class DocManager:
    def __init__(self, dbt_doc_resolver: DBTDocResolver) -> None:
        self._dbt_doc_resolver = dbt_doc_resolver
        self._dbt_model_doc_store: Dict[str, DBTModelDocument] = {}
        self._dbt_model_sql_doc_store: Dict[str, DBTModelSqlDocument] = {}

        self._initialize_model_doc_store()

    def _get_catalog_columns_for_unique_id(self, unique_id: str) -> List[CatalogColumn]:
        model = self._dbt_doc_resolver.get_manifest_by_unique_id(unique_id) or {
            "columns": {}
        }
        _columns: Dict[str, Dict[str, str]] = defaultdict(dict)
        for column in model["columns"].values():
            _columns[column["name"]]["description"] = column["description"]
            _columns[column["name"]]["type"] = column["data_type"]
            _columns[column["name"]]["name"] = column["name"]

        for column in (
            (self._dbt_doc_resolver.get_catalog_by_unique_id(unique_id) or {})
            .get("columns", {})
            .values()
        ):
            _columns[column["name"]]["type"] = column["type"]
            _columns[column["name"]]["name"] = column["name"]

        catalog_columns = [
            CatalogColumn(
                data_type=i["type"],
                description=i.get("description"),
                name=i["name"],
            )
            for i in _columns.values()
        ]
        return catalog_columns

    def _initialize_model_doc_store(self):
        for unique_id in self._dbt_doc_resolver.list_model_unique_id():
            model = self._dbt_doc_resolver.get_manifest_by_unique_id(unique_id)
            if model["resource_type"] == "model":
                model_name = _truncate_schema_name_for_model(model["unique_id"])
                unique_id = model["unique_id"]
                catalog_columns = self._get_catalog_columns_for_unique_id(unique_id)

                model_doc = DBTModelDocument(
                    name=model_name,
                    description=model["description"],
                    columns=catalog_columns,
                    depends_on=[
                        _truncate_schema_name_for_model(i)
                        for i in model["depends_on"]["nodes"]
                    ],
                    meta=DocMetaContainer(
                        doc_type=DocType.MODEL, meta=DBTDocMeta(name=model_name)
                    ),
                )
                self._dbt_model_doc_store[model_name] = model_doc

                if "compiled_code" in model:
                    model_sql_doc = DBTModelSqlDocument(
                        doc=model_doc,
                        compiled_sql=model["compiled_code"],
                        meta=DocMetaContainer(
                            doc_type=DocType.MODEL, meta=DBTDocMeta(name=model_name)
                        ),
                    )
                    self._dbt_model_sql_doc_store[model_name] = model_sql_doc

    def get_all_docs(self) -> List[Doc]:
        """Get all DBT docs"""
        logging.debug("Getting all DBT docs")
        return [
            *self._dbt_model_doc_store.values(),
        ]

    def _resolve_dbt_doc_meta(self, container: DocMetaContainer) -> Doc:
        """Resolve a DBT doc meta"""
        model_meta = container.meta
        if container.doc_type == DocType.MODEL:
            return self._dbt_model_doc_store[model_meta["name"]]
        elif container.doc_type == DocType.SQL:
            return self._dbt_model_sql_doc_store[model_meta["name"]]
        else:
            raise NotImplementedError

    def resolve_doc_meta(self, container: DocMetaContainer) -> Doc:
        if container.doc_type in [
            DocType.MODEL,
            DocType.SQL,
        ]:
            return self._resolve_dbt_doc_meta(container)
        elif container.doc_type == DocType.CHAT:
            meta = ChatMessageMeta.parse_obj(container.meta)
            return ChatConversationDocument(
                query=meta.query,
                response=meta.response,
                meta=container,
                ref_dbt_docs=[
                    self._resolve_dbt_doc_meta(i)
                    for i in meta.ref_dbt_docs_meta_containers
                ],
            )
        else:
            raise NotImplementedError


class ChatBot:
    """Chatbot for dbt documentation"""

    def __init__(
        self,
        doc_resolver: DBTDocResolver,
        vector_storage: VectorStorage,
        tiktoken_provider: Optional[TikTokenProvider],
        openai_config: Optional[Dict[str, Any]] = None,
        i18n: str = "en",
    ) -> None:
        self.doc_manager = DocManager(doc_resolver)
        self.vector_storage = vector_storage
        self.tiktoken_provider = tiktoken_provider
        self.openai = Openai(**(openai_config or {}))
        self._i18n = i18n
        self._messages: List[ChatMessage] = []

    def index_dbt_docs(self):
        """Index all dbt docs"""

        if self.tiktoken_provider:
            n_tokens = 0
            for doc in self.doc_manager.get_all_docs():
                n_tokens += self.tiktoken_provider.count_token(
                    doc.get_content(), EMBEDDING_MODEL
                )
            logging.info(
                "index dbt docs total tokens: %s, cost %s$",
                n_tokens,
                price_for_embedding(n_tokens),
            )
        for doc in self.doc_manager.get_all_docs():
            logging.debug("indexing doc: %s", doc)
            vector = self.openai.embed(doc.get_content())
            self.vector_storage.insert_doc(doc, vector)

    def _unique_docs(self, docs: List[Doc]) -> List[Doc]:
        """Remove duplicate docs"""
        res = []
        visited: Set[str] = set()
        for doc in docs:
            if doc.get_unique_id() not in visited:
                res.append(doc)
                visited.add(doc.get_unique_id())
        return res

    def suggest_table(self, query: str, k: int = 5) -> ChatMessage:
        """Suggest table for query"""
        vector = self.openai.embed(query)
        logging.debug("embedding query: %s, %s", query, vector[:5])
        similar_docs_meta = self.vector_storage.similarity_search(vector, k)
        docs = [self.doc_manager.resolve_doc_meta(meta) for meta in similar_docs_meta]
        logging.debug("similar docs: %s, %s", query, docs)
        dbt_docs = [
            doc
            for doc in docs
            if doc.get_metadata().doc_type in [DocType.MODEL, DocType.SQL]
        ]
        chat_docs = [
            cast(ChatConversationDocument, doc)
            for doc in docs
            if doc.get_metadata().doc_type == DocType.CHAT
        ]
        dbt_model_names = [doc.get_metadata().meta["name"] for doc in dbt_docs]

        if not dbt_docs:
            message = ChatMessage(
                uuid=uuid.uuid4().hex,
                created_at=datetime.datetime.now(),
                query=query,
                response=get_i18n_text(I18nKey.KEY_NO_RESPONSE),
                ref_dbt_docs=dbt_docs,
                ref_chat_docs=chat_docs,
            )
            self._messages.append(message)
            return message

        messages = []
        messages.append(
            {
                "role": "system",
                "content": get_i18n_text(I18nKey.KEY_PROMPT_SYSTEM_ROLE_CONTENT),
            }
        )
        if dbt_docs:
            _content = "\n".join([doc.get_content() for doc in docs])
            messages.append(
                {
                    "role": "system",
                    "content": get_i18n_text(
                        I18nKey.KEY_PROMPT_SYSTEM_ROLE_RELATED_TABLES
                    ).format(_content),
                }
            )

        messages.append(
            {
                "role": "user",
                "content": get_i18n_text(
                    I18nKey.KEY_PROMPT_USER_ROLE_SUGGEST_TABLES
                ).format(query, ",".join(dbt_model_names)),
            }
        )
        response = self.openai.chat_completion(messages=messages)
        message = ChatMessage(
            uuid=uuid.uuid4().hex,
            created_at=datetime.datetime.now(),
            query=query,
            response=response,
            ref_dbt_docs=dbt_docs,
            ref_chat_docs=chat_docs,
        )
        self._messages.append(message)
        return message

    def suggest_sql(self, query: str, k: int = 10) -> ChatMessage:
        """Suggest sql for query"""
        vector = self.openai.embed(query)
        logging.debug("embedding query: %s, %s", query, vector[:5])
        similar_docs_meta = self.vector_storage.similarity_search(vector, k)
        docs = [self.doc_manager.resolve_doc_meta(meta) for meta in similar_docs_meta]
        logging.debug("similar docs: %s, %s", query, docs)
        dbt_docs = [
            doc
            for doc in docs
            if doc.get_metadata().doc_type in [DocType.MODEL, DocType.SQL]
        ]
        chat_docs = [
            cast(ChatConversationDocument, doc)
            for doc in docs
            if doc.get_metadata().doc_type == DocType.CHAT
        ]
        dbt_model_names = [doc.get_metadata().meta["name"] for doc in dbt_docs]

        if not dbt_docs:
            message = ChatMessage(
                uuid=uuid.uuid4().hex,
                created_at=datetime.datetime.now(),
                query=query,
                response=get_i18n_text(I18nKey.KEY_NO_RESPONSE),
                ref_dbt_docs=dbt_docs,
                ref_chat_docs=chat_docs,
            )
            self._messages.append(message)
            return message

        messages = []
        messages.append(
            {
                "role": "system",
                "content": get_i18n_text(I18nKey.KEY_PROMPT_SYSTEM_ROLE_CONTENT),
            }
        )
        if dbt_docs:
            _content = "\n".join([doc.get_content() for doc in docs])
            messages.append(
                {
                    "role": "system",
                    "content": get_i18n_text(
                        I18nKey.KEY_PROMPT_SYSTEM_ROLE_RELATED_TABLES
                    ).format(_content),
                }
            )

        messages.append(
            {
                "role": "user",
                "content": get_i18n_text(
                    I18nKey.KEY_PROMPT_USER_ROLE_SUGGEST_SQL
                ).format(query, ",".join(dbt_model_names)),
            }
        )
        logging.debug("messages: %s", messages)
        response = self.openai.chat_completion(messages=messages)
        message = ChatMessage(
            uuid=uuid.uuid4().hex,
            created_at=datetime.datetime.now(),
            query=query,
            response=response,
            ref_dbt_docs=dbt_docs,
            ref_chat_docs=chat_docs,
        )
        self._messages.append(message)
        return message

    def memory_message(self, message: ChatMessage):
        """Memory message"""
        return self.memory_message_by_uuid(message.uuid)

    def _find_message(self, chat_uuid: str) -> Optional[ChatMessage]:
        """Find message by uuid"""
        for message in self._messages:
            if message.uuid == chat_uuid:
                return message
        return None

    def memory_message_by_uuid(self, chat_uuid: str):
        """Memory message by uuid"""
        message = self._find_message(chat_uuid)
        if not message:
            raise ValueError("message not found")
        chat_doc = ChatConversationDocument(
            query=message.query,
            response=message.response,
            meta=DocMetaContainer(
                doc_type=DocType.CHAT,
                meta=ChatMessageMeta(
                    query=message.query,
                    response=message.response,
                    created_at=datetime.datetime.now(),
                    ref_dbt_docs_meta_containers=[
                        i.get_metadata() for i in message.ref_dbt_docs
                    ],
                ).dict(),
            ),
            ref_dbt_docs=message.ref_dbt_docs,
        )

        vector = self.openai.embed(chat_doc.get_content())
        logging.debug("memory message: %s, %s", chat_doc, vector[:5])
        self.vector_storage.insert_doc(chat_doc, vector)



import logging
from typing import List
from chatdbt.model import DocMetaContainer, VectorStorage, Doc


class AtlasVectorStorage(VectorStorage):
    """Vector storage using Atlas"""

    def __init__(self, api_key: str, project_name: str):
        self.api_key = api_key
        self.project_name = project_name
        try:
            from nomic import AtlasProject  # pylint: disable=import-outside-toplevel
        except ImportError as ex:
            raise RuntimeError(
                f"{ex}\nPlease install nomic to use AtlasVectorStorage\npip install chatdbt[nomic]"
            )  # pylint: disable=raise-missing-from

        logging.debug("initializing Atlas project: %s", project_name)
        self.project = AtlasProject(
            name=project_name,
            modality="embedding",
            is_public=True,
            unique_id_field="foobar",
        )
        logging.debug("Atlas project: %s", self.project)

    def insert_doc(self, doc: Doc, vector: List[float]):
        """Insert a document into the vector storage"""
        with self.project.wait_for_project_lock():
            self.project.add_embeddings(
                data=[doc.get_metadata().dict()], embeddings=[vector]
            )

    def similarity_search(self, vector: List[float], k: int) -> List[DocMetaContainer]:
        """Search for similar documents in the vector storage"""
        with self.project.wait_for_project_lock():
            neighbors, _ = self.project.projections[0].vector_search([vector], k=k)
            datas = self.project.get_data(ids=neighbors)

        res = [DocMetaContainer(**datas[idx]) for idx in enumerate(neighbors)]
        return res



"""Base classes for the chatdbt model"""
import datetime
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional
from pydantic import BaseModel as PydanticBaseModel
from chatdbt.i18n import get_i18n_text, I18nKey

from enum import Enum


class BaseModel(PydanticBaseModel):
    class Config:
        arbitrary_types_allowed = True


class DocType(Enum):
    MODEL = "model"
    # todo add metrics
    # METRICS = "metrics"
    SQL = "sql"
    CHAT = "chat"


class DocMetaContainer(BaseModel):
    """Metadata for a document"""

    doc_type: DocType
    meta: Dict[str, Any]


class Doc(ABC):
    """Base class for all documents"""

    @abstractmethod
    def get_content(self):
        """Get the content of the document"""

    @abstractmethod
    def get_metadata(self) -> DocMetaContainer:
        """Get the metadata of the document"""

    @abstractmethod
    def get_unique_id(self) -> str:
        pass


class VectorStorage(ABC):
    """Base class for all vector storages"""

    @abstractmethod
    def insert_doc(self, doc: Doc, vector: List[float]):
        """Insert a document into the vector storage"""

    @abstractmethod
    def similarity_search(self, vector: List[float], k: int) -> List[DocMetaContainer]:
        """Search for similar documents in the vector storage"""


class EmbeddingProvider(ABC):
    """Base class for all embeddings"""

    @abstractmethod
    def embed(self, content: str) -> List[float]:
        """Embed a piece of text into a vector"""


class TikTokenProvider(ABC):
    """Base class for all tiktoken providers"""

    @abstractmethod
    def count_token(self, prompt: str, model: str) -> Optional[int]:
        """Count tokens for a prompt"""


def _markdown_dbt_doc_li(name: str, content: str, indent: int = 4):
    _items = content.split("\n")
    items = []
    for item in _items:
        items.append(item)
    body = "\n\n".join([" " * indent + item for item in ["```", *items, "```"]])
    return f"""- {name}

{body}

"""


def _markdown_chat_doc_li(query: str, response: str, indent: int = 4):
    items = [
        " " * indent + item
        for item in ["```", f"query: {query}", f"response: {response}", "```"]
    ]
    body = "\n\n".join(items)
    return f"- {body}"


class DBTDocResolver(ABC):
    """Base class for all DBT document resolvers"""

    @abstractmethod
    def get_manifest_by_unique_id(self, unique_id: str) -> Optional[Dict]:
        pass

    @abstractmethod
    def get_catalog_by_unique_id(self, unique_id: str) -> Optional[Dict]:
        pass

    @abstractmethod
    def list_model_unique_id(self) -> List[str]:
        pass


class DBTDocMeta(BaseModel):
    """DBT document metadata"""

    name: str


class ChatMessageMeta(BaseModel):
    """Chat message metadata"""

    ref_dbt_docs_meta_containers: List[DocMetaContainer]
    query: str
    created_at: datetime.datetime
    response: Optional[str]


class CatalogColumn(BaseModel):
    """A column in a DBT catalog"""

    name: str
    data_type: Optional[str]
    description: Optional[str]


class DBTModelDocument(BaseModel, Doc):
    """A DBT model document"""

    meta: DocMetaContainer
    name: str
    description: Optional[str]
    columns: List[CatalogColumn]
    depends_on: List[str]

    def get_content(self):
        columns = [
            str(i.dict()).replace(" ", "").replace("'", "") for i in self.columns
        ]
        return f"""name: {self.name}
description: {self.description or ''}
columns: {columns}
depends_on: {self.depends_on}
"""

    def get_metadata(self):
        return self.meta

    def get_unique_id(self) -> str:
        return self.name


# todo: add metrics document


class DBTModelSqlDocument(BaseModel, Doc):
    """A DBT SQL document"""

    doc: DBTModelDocument
    compiled_sql: str
    meta: DocMetaContainer

    def get_content(self):
        return f"""name: {self.doc.name}
build_by: {self.compiled_sql}
depends_on: {self.doc.depends_on}
"""

    def get_metadata(self):
        return self.meta

    def get_unique_id(self) -> str:
        return f"{self.doc.name}_sql"


class ChatConversationDocument(BaseModel, Doc):
    """A chat conversation document"""

    query: str
    response: Optional[str]
    meta: DocMetaContainer
    ref_dbt_docs: List[Doc]

    def get_content(self):
        return f"""previous chat:
query: {self.query}
response: {self.response}"""

    def get_metadata(self):
        return self.meta

    def get_unique_id(self) -> str:
        return str(hash(self.query))


class ChatMessage(BaseModel):
    """A chat message"""

    query: str
    response: str
    uuid: str
    created_at: datetime.datetime
    ref_dbt_docs: List[Doc]
    ref_chat_docs: List[ChatConversationDocument]

    def _repr_markdown_(self):
        ref_dbt_docs = "\n".join(
            [
                _markdown_dbt_doc_li(i.get_unique_id(), i.get_content())
                for i in self.ref_dbt_docs
            ]
        )
        ref_chat_docs = "\n".join(
            [
                _markdown_chat_doc_li(
                    i.query,
                    i.response,
                )
                for i in self.ref_chat_docs
            ]
        )

        return f"""{self.response}

{get_i18n_text(I18nKey.KEY_MESSAGE_related_tables)}:

{ref_dbt_docs}

{get_i18n_text(I18nKey.KEY_MESSAGE_related_messages)}:

{ref_chat_docs}

"""



import functools
import logging
import os
from typing import Optional, cast, Any, Dict

from chatdbt.chat import ChatBot
from chatdbt.dbt_doc_resolver import get_dbt_doc_resolver
from chatdbt.model import ChatMessage, DBTDocResolver, TikTokenProvider, VectorStorage
from chatdbt.tiktoken_provider import get_tiktoken_provider
from chatdbt.vector_storage import get_vector_storage


ENV_VAR_VECTOR_STORAGE_TYPE = "CHATDBT_VECTOR_STORAGE_TYPE"
ENV_VAR_VECTOR_STORAGE_CONFIG_PREFIX = "CHATDBT_VECTOR_STORAGE_CONFIG_"
ENV_VAR_DBT_DOC_RESOLVER_TYPE = "CHATDBT_DBT_DOC_RESOLVER_TYPE"
ENV_VAR_DBT_DOC_RESOLVER_CONFIG_PREFIX = "CHATDBT_DBT_DOC_RESOLVER_CONFIG_"

ENV_VAR_I18N = "CHATDBT_I18N"

ENV_VAR_TIKTOKEN_PROVIDER_TYPE = "CHATDBT_TIKTOKEN_PROVIDER_TYPE"
ENV_VAR_TIKTOKEN_PROVIDER_CONFIG_PREFIX = "CHATDBT_TIKTOKEN_PROVIDER_CONFIG_"

ENV_VAR_OPENAI_CONFIG_PREFIX = "CHATDBT_OPENAI_CONFIG_"


class _Global:
    chat_instance: Optional[ChatBot] = None
    chat_instance_init: bool = False


def setup_shortcut(
    vector_storage: VectorStorage,
    dbt_doc_resolver: DBTDocResolver,
    tiktoken_provider: Optional[TikTokenProvider] = None,
    openai_config: Optional[Dict[str, Any]] = None,
    i18n: str = "en-us",
):
    logging.basicConfig(level=logging.INFO)

    _Global.chat_instance = ChatBot(
        dbt_doc_resolver,
        vector_storage,
        tiktoken_provider,
        openai_config,
        i18n,
    )
    _Global.chat_instance_init = True


def setup_shortcut_via_env() -> None:
    """Setup chat instance via environment variables"""
    vector_storage_type = os.environ.get(ENV_VAR_VECTOR_STORAGE_TYPE)
    if not vector_storage_type:
        raise ValueError(
            f"Environment variable {ENV_VAR_VECTOR_STORAGE_TYPE} is not set"
        )
    vector_storage_config = {
        k.replace(ENV_VAR_VECTOR_STORAGE_CONFIG_PREFIX, "").lower(): v
        for k, v in os.environ.items()
        if k.startswith(ENV_VAR_VECTOR_STORAGE_CONFIG_PREFIX)
    }
    dbt_doc_resolver_type = os.environ.get(ENV_VAR_DBT_DOC_RESOLVER_TYPE)
    if not dbt_doc_resolver_type:
        raise ValueError(
            f"Environment variable {ENV_VAR_DBT_DOC_RESOLVER_TYPE} is not set"
        )
    dbt_doc_resolver_config = {
        k.replace(ENV_VAR_DBT_DOC_RESOLVER_CONFIG_PREFIX, "").lower(): v
        for k, v in os.environ.items()
        if k.startswith(ENV_VAR_DBT_DOC_RESOLVER_CONFIG_PREFIX)
    }

    i18n = os.environ.get(ENV_VAR_I18N, "en")

    tiktoken_provider: Optional[TikTokenProvider] = None
    tiktoken_provider_type = os.environ.get(ENV_VAR_TIKTOKEN_PROVIDER_TYPE)
    tiktoken_provider_config = {
        k.replace(ENV_VAR_TIKTOKEN_PROVIDER_CONFIG_PREFIX, "").lower(): v
        for k, v in os.environ.items()
        if k.startswith(ENV_VAR_TIKTOKEN_PROVIDER_CONFIG_PREFIX)
    }

    if tiktoken_provider_type is not None:
        tiktoken_provider = get_tiktoken_provider(
            tiktoken_provider_type, tiktoken_provider_config
        )

    openai_config = {
        k.replace(ENV_VAR_OPENAI_CONFIG_PREFIX, "").lower(): v
        for k, v in os.environ.items()
        if k.startswith(ENV_VAR_OPENAI_CONFIG_PREFIX)
    }

    setup_shortcut(
        get_vector_storage(vector_storage_type, vector_storage_config),
        get_dbt_doc_resolver(dbt_doc_resolver_type, dbt_doc_resolver_config),
        tiktoken_provider,
        openai_config,
        i18n,
    )


def ensure_chat_init(func):
    """Decorator to ensure chat instance is initialized before calling the function."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not _Global.chat_instance_init:
            setup_shortcut_via_env()
        return func(*args, **kwargs)

    return wrapper


@ensure_chat_init
def suggest_table(query: str, k: int = 5):
    """Suggest table based on query."""
    chat: ChatBot = cast(ChatBot, _Global.chat_instance)
    return chat.suggest_table(query, k)


@ensure_chat_init
def index_dbt_docs() -> None:
    """Index dbt docs."""
    chat: ChatBot = cast(ChatBot, _Global.chat_instance)
    return chat.index_dbt_docs()


@ensure_chat_init
def suggest_sql(query: str, k: int = 5):
    """Suggest sql based on query."""
    chat: ChatBot = cast(ChatBot, _Global.chat_instance)
    return chat.suggest_sql(query, k)


@ensure_chat_init
def memory_message(message: ChatMessage):
    """Memory chat message."""
    chat: ChatBot = cast(ChatBot, _Global.chat_instance)
    return chat.memory_message(message)



__version__ = "0.1.0"

from .chat import ChatBot
from .shortcut import index_dbt_docs, memory_message, suggest_sql, suggest_table


__all__ = [
    "suggest_sql",
    "suggest_table",
    "memory_message",
    "index_dbt_docs",
    "ChatBot",
]



from typing import Dict, Any
from chatdbt.model import VectorStorage


def get_vector_storage(
    vector_storage_type: str, vector_storage_config: Dict[str, Any]
) -> VectorStorage:
    """Get a vector storage instance"""
    if vector_storage_type == "atlas":
        from chatdbt.vector_storage.atlas import AtlasVectorStorage

        return AtlasVectorStorage(**vector_storage_config)
    elif vector_storage_type == "pgvector":
        from chatdbt.vector_storage.pgvector import PGVectorStorage

        return PGVectorStorage(**vector_storage_config)
    else:
        raise ValueError("Unknown vector storage type")



import logging
from typing import Optional

import requests

from chatdbt.model import TikTokenProvider


class TikTokenHttpServerProvider(TikTokenProvider):
    """Count token using a"""

    def __init__(self, api_base: str, timeout_secs: int = 5) -> None:
        self._api_base = api_base.rstrip("/")
        self._timeout_secs = timeout_secs

    def count_token(self, prompt: str, model: str) -> Optional[int]:
        """Count tokens for a prompt"""

        req = requests.get(
            url=f"{self._api_base}/tokenize",
            json={"prompt": prompt, "model": model},
            timeout=self._timeout_secs,
        )
        if not req.ok:
            logging.warning(
                "Failed to count token, got %s %s", req.status_code, req.text
            )
            return None
        tokens = req.json()["tokens"]
        return len(tokens)



from typing import Any, Dict

from chatdbt.model import TikTokenProvider


def get_tiktoken_provider(
    provider_type: str, config: Dict[str, Any]
) -> TikTokenProvider:
    if provider_type == "tiktoken_http_server":
        from .tiktoken_http_server import TikTokenHttpServerProvider

        return TikTokenHttpServerProvider(**config)
    else:
        raise NotImplementedError



# ruff: noqa: E501
import os
from enum import Enum


class I18nKey(Enum):
    KEY_PROMPT_SYSTEM_ROLE_CONTENT = "prompt_system_content"
    KEY_PROMPT_SYSTEM_ROLE_RELATED_TABLES = "prompt_system_related_tables"
    KEY_PROMPT_SYSTEM_ROLE_RELATED_MESSAGES = "prompt_system_related_messages"
    KEY_PROMPT_USER_ROLE_SUGGEST_TABLES = "prompt_user_suggest_tables"
    KEY_PROMPT_USER_ROLE_SUGGEST_SQL = "prompt_user_suggest_sql"
    KEY_NO_RESPONSE = "no_response"

    KEY_MESSAGE_related_tables = "message_related_tables"
    KEY_MESSAGE_related_messages = "message_related_messages"


_i18n = {
    "en-us": {
        I18nKey.KEY_PROMPT_SYSTEM_ROLE_CONTENT: "You are a sql master",
        I18nKey.KEY_PROMPT_SYSTEM_ROLE_RELATED_TABLES: "Here are some tables with description: {}",
        I18nKey.KEY_PROMPT_SYSTEM_ROLE_RELATED_MESSAGES: "Here are some related chat messages: {}",
        I18nKey.KEY_PROMPT_USER_ROLE_SUGGEST_TABLES: "My question is '{}'. Can you tell me which table in ({}) I should use? why?",
        I18nKey.KEY_PROMPT_USER_ROLE_SUGGEST_SQL: "My question is '{}'. Can you tell me which table in ({}) I should use? why? Can you write a sql to complete my question based on this table?",
        I18nKey.KEY_NO_RESPONSE: "Sorry, I can't find any tables",
        I18nKey.KEY_MESSAGE_related_tables: "Related tables",
        I18nKey.KEY_MESSAGE_related_messages: "Related messages",
    },
    "zh-cn": {
        I18nKey.KEY_PROMPT_SYSTEM_ROLE_CONTENT: "你是一位 SQL 专家",
        I18nKey.KEY_PROMPT_SYSTEM_ROLE_RELATED_TABLES: "这里有一些 table 的描述: {}",
        I18nKey.KEY_PROMPT_SYSTEM_ROLE_RELATED_MESSAGES: "这里有一些相关的聊天记录: {}",
        I18nKey.KEY_PROMPT_USER_ROLE_SUGGEST_TABLES: "我的问题是 '{}'. 你能告诉我我应该使用哪个 table 吗? 为什么?",
        I18nKey.KEY_PROMPT_USER_ROLE_SUGGEST_SQL: "我的问题是 '{}'. 你能告诉我我应该使用哪个 table 吗? 为什么? 你能基于这个 table ，写一段 sql 来完成我的问题吗？",
        I18nKey.KEY_NO_RESPONSE: "抱歉，我检索不到任何 table",
        I18nKey.KEY_MESSAGE_related_tables: "相关的 table",
        I18nKey.KEY_MESSAGE_related_messages: "相关的聊天记录",
    },
}


def get_i18n_text(key: I18nKey) -> str:
    lang = os.environ.get("CHATDBT_I18N", "en-us")
    if lang not in _i18n:
        raise ValueError(f"Unsupported language: {lang}")
    return _i18n[lang][key]



import json
from typing import Dict, List, Optional

from chatdbt.model import DBTDocResolver


class LocalfsDBTDocResolver(DBTDocResolver):
    def __init__(self, manifest_json_path: str, catalog_json_path: str):
        with open(manifest_json_path, "r", encoding="utf8") as manifest_f:
            self._dbt_docs_manifest = json.load(manifest_f)
        with open(catalog_json_path, "r", encoding="utf8") as catalog_f:
            self._dbt_docs_catalog = json.load(catalog_f)

    def get_manifest_by_unique_id(self, unique_id: str) -> Optional[Dict]:
        return self._dbt_docs_manifest["nodes"].get(unique_id)

    def get_catalog_by_unique_id(self, unique_id: str) -> Optional[Dict]:
        return self._dbt_docs_catalog["nodes"].get(unique_id)

    def list_model_unique_id(self) -> List[str]:
        res = []
        for model in self._dbt_docs_manifest["nodes"].values():
            if model["resource_type"] != "model":
                continue
            res.append(model["unique_id"])
        return res



import json
import logging
import datetime
from sqlalchemy import Column, create_engine, Integer, VARCHAR, DateTime
from sqlalchemy.dialects.postgresql import insert
from pgvector.sqlalchemy import Vector
from sqlalchemy.dialects.postgresql import JSON
from sqlalchemy.orm import Session, declarative_base
from chatdbt.model import VectorStorage, Doc, DocMetaContainer

from typing import List


Base = declarative_base()


class PGVectorStorage(VectorStorage):
    def _orm_for(self, table_name):
        class Item(Base):
            __tablename__ = table_name
            id = Column(Integer, primary_key=True, autoincrement=True)
            unique_id = Column(VARCHAR, unique=True)
            embedding = Column(Vector(1536))
            data_metadata = Column(JSON)
            created_at = Column(DateTime, default=datetime.datetime.utcnow)
            updated_at = Column(
                DateTime,
                default=datetime.datetime.utcnow,
                onupdate=datetime.datetime.utcnow,
            )

        return Item

    def __init__(self, connect_string: str, table_name: str):
        self.connect_string = connect_string
        self.table_name = table_name
        self._engine = create_engine(connect_string)
        self._conn = self._engine.connect()

        self._table = self._orm_for(table_name)
        self._create_tables()

    def _create_tables(self):
        Base.metadata.create_all(self._engine)

    def insert_doc(self, doc: Doc, vector: List[float]):
        logging.debug("inserting doc: %s, %s", doc, vector[:5])
        unique_id = doc.get_unique_id()
        with Session(self._conn) as session:
            values = dict(
                unique_id=unique_id,
                embedding=vector,
                data_metadata=doc.get_metadata().json(),
                updated_at=datetime.datetime.utcnow(),
            )

            stmt = (
                insert(self._table)
                .values(
                    **values,
                )
                .on_conflict_do_update(
                    index_elements=["unique_id"],
                    set_=values,
                )
            )
            session.execute(stmt)
            session.commit()

    def similarity_search(self, vector: List[float], k: int) -> List[DocMetaContainer]:
        with Session(self._conn) as session:
            res = (
                session.query(self._table)
                .order_by(self._table.embedding.cosine_distance(vector))
                .limit(k)
                .all()
            )
        return [
            DocMetaContainer.parse_obj(json.loads(item.data_metadata)) for item in res
        ]



from typing import Dict
from chatdbt.model import DBTDocResolver


def get_dbt_doc_resolver(
    dbt_doc_resolver_type: str, dbt_doc_resolver_config: Dict[str, str]
) -> DBTDocResolver:
    if dbt_doc_resolver_type == "localfs":
        from .localfs import LocalfsDBTDocResolver

        return LocalfsDBTDocResolver(**dbt_doc_resolver_config)
    else:
        raise ValueError("Unknown dbt doc resolver type")
